from xml.etree.ElementTree import Element, SubElement, tostring

import requests
from defusedxml.ElementTree import fromstring
from moneyed import Money
from structlog import get_logger

from .xml_helpers import money_to_amount_and_currency, parse_money
from ..config import datatrans_processor_url, sign_web
from ..models import Payment, Refund

logger = get_logger()


def refund(amount: Money, payment_id: str) -> Refund:
    """
    Refunds (partially or completely) a previously authorized and settled payment.
    :param amount: The amount and currency we want to refund. Must be positive, in the same currency
    as the original payment, and not exceed the amount of the original payment.
    :param payment_id: The id of the payment to refund.
    :return: a Refund (either successful or not).
    """
    if amount.amount <= 0:
        raise ValueError('Refund takes a strictly positive amount')
    payment = Payment.objects.get(pk=payment_id)
    if not payment.success:
        raise ValueError('Only successful payments can be refunded')
    if payment.amount.currency != amount.currency:
        raise ValueError('Refund currency must be identical to original payment currency')
    if amount.amount > payment.amount.amount:
        raise ValueError('Refund amount exceeds original payment amount')

    logger.info('refunding-payment', amount=str(amount),
                payment=dict(amount=str(payment.amount), transaction_id=payment.transaction_id,
                             masked_card_number=payment.masked_card_number))

    client_ref = '{}-r'.format(payment.client_ref)

    request_xml = build_refund_request_xml(amount=amount,
                                           original_transaction_id=payment.transaction_id,
                                           client_ref=client_ref,
                                           merchant_id=payment.merchant_id)

    logger.info('sending-refund-request', url=datatrans_processor_url, data=request_xml)

    response = requests.post(
        url=datatrans_processor_url,
        headers={'Content-Type': 'application/xml'},
        data=request_xml)

    logger.info('processing-refund-response', response=response.content)

    refund_response = parse_refund_response_xml(response.content)
    refund_response.save()
    refund_response.send_signal()

    return refund_response


def build_refund_request_xml(amount: Money, client_ref: str, original_transaction_id: str, merchant_id: str) -> bytes:
    client_ref = client_ref

    root = Element('paymentService')
    root.set('version', '1')

    body = SubElement(root, 'body')
    body.set('merchantId', merchant_id)

    transaction = SubElement(body, 'transaction')
    transaction.set('refno', client_ref)

    request = SubElement(transaction, 'request')

    amount, currency = money_to_amount_and_currency(amount)
    SubElement(request, 'amount').text = str(amount)
    SubElement(request, 'currency').text = currency
    SubElement(request, 'uppTransactionId').text = original_transaction_id
    SubElement(request, 'transtype').text = '06'
    SubElement(request, 'sign').text = sign_web(merchant_id, amount, currency, client_ref)

    return tostring(root, encoding='utf8')


def parse_refund_response_xml(xml: bytes) -> Refund:
    body = fromstring(xml).find('body')
    status = body.get('status')
    transaction = body.find('transaction')
    trx_status = transaction.get('trxStatus')
    request = transaction.find('request')

    common_attributes = dict(
        merchant_id=body.get('merchantId'),
        client_ref=transaction.get('refno'),
        amount=parse_money(request),
        payment_transaction_id=request.find('uppTransactionId').text,
    )

    request_type = request.find('reqtype')
    if request_type is not None:
        common_attributes['request_type'] = request_type.text

    if status == 'accepted' and trx_status == 'response':
        response = transaction.find('response')

        transaction_id = response.find('uppTransactionId').text
        response_code = response.find('responseCode').text
        response_message = response.find('responseMessage').text
        authorization_code = response.find('authorizationCode').text
        acquirer_authorization_code = response.find('acqAuthorizationCode').text

        d = dict(
            success=True,
            transaction_id=transaction_id,
            response_code=response_code,
            response_message=response_message,
            authorization_code=authorization_code,
            acquirer_authorization_code=acquirer_authorization_code,
        )
        d.update(common_attributes)
        return Refund(**d)
    else:
        error = transaction.find('error')

        acquirer_error_code_element = error.find('acqErrorCode')

        d = dict(
            success=False,
            error_code=error.find('errorCode').text,
            error_message=error.find('errorMessage').text,
            error_detail=error.find('errorDetail').text,
            acquirer_error_code=acquirer_error_code_element.text if acquirer_error_code_element is not None else '',
        )
        d.update(**common_attributes)
        return Refund(**d)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-30 05:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('datatrans', '0003_add_modified_field'),
    ]

    operations = [
        migrations.AlterField(
            model_name='aliasregistration',
            name='card_alias',
            field=models.CharField(blank=True, db_index=True, max_length=20),
        ),
        migrations.AlterField(
            model_name='aliasregistration',
            name='masked_card_number',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



from django.conf.urls import url
from django.contrib import admin
from django.forms import CharField, TextInput, forms
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.html import format_html
from djmoney.forms import MoneyField
from moneyed.localization import format_money

from .gateway import pay_with_alias, refund
from .models import AliasRegistration, Payment, Refund


def expiry(obj):
    if obj.expiry_month is not None and obj.expiry_year is not None:
        return format_html('{}/{}', obj.expiry_month, obj.expiry_year)


def amount(obj):
    return format_money(obj.amount)


class PayWithAliasForm(forms.Form):
    amount = MoneyField(min_value=0, default_currency='CHF')
    client_ref = CharField(required=True, max_length=18, widget=TextInput(attrs={'size': 18}))


def pay_with_alias_form(request, alias_registration_id):
    if request.method == 'POST':
        form = PayWithAliasForm(request.POST)
        if form.is_valid():
            result = pay_with_alias(
                amount=form.cleaned_data['amount'],
                alias_registration_id=alias_registration_id,
                client_ref=form.cleaned_data['client_ref'],
            )
            # As confirmation we take the user to the detail page of the payment.
            return HttpResponseRedirect(reverse('admin:datatrans_payment_change', args=[result.id]))
    else:
        form = PayWithAliasForm()

    alias_registration = get_object_or_404(AliasRegistration, pk=alias_registration_id)

    return render(
        request,
        'admin/datatrans/form.html',
        {
            'title': 'Pay with alias of credit card {}'.format(alias_registration.masked_card_number),
            'form': form,
            'opts': AliasRegistration._meta,  # Used to setup the navigation / breadcrumbs of the page
        }
    )


@admin.register(AliasRegistration)
class AliasRegistrationAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    readonly_fields = ['created', 'modified', 'expiry_date', 'pay_with_alias_button']
    list_display = [
        'transaction_id', 'created', 'success', 'client_ref', amount, 'payment_method', 'card_alias',
        'masked_card_number', expiry,
        'credit_card_country', 'error_code', 'pay_with_alias_button']
    search_fields = [
        'transaction_id', 'created', 'expiry_date', 'payment_method', 'card_alias', 'masked_card_number',
        'expiry_month', 'expiry_year', 'credit_card_country', 'client_ref', 'response_code',
        'authorization_code', 'acquirer_authorization_code', 'error_code', 'error_message', 'error_detail']
    list_filter = ['success', 'payment_method', 'credit_card_country']
    ordering = ['-created']

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            url(
                r'^(?P<alias_registration_id>[0-9a-f-]+)/pay_with_alias/$',
                self.admin_site.admin_view(pay_with_alias_form),
                name='datatrans_alias_pay',
            ),
        ]
        return my_urls + urls

    def pay_with_alias_button(self, obj):
        if obj.success:
            return format_html('<a class="button" href="{}">Pay with alias</a>',
                               reverse('admin:datatrans_alias_pay', args=[obj.pk]))
        else:
            return '-'

    pay_with_alias_button.short_description = 'Pay with alias'  # type: ignore


class RefundPaymentForm(forms.Form):
    amount = MoneyField(min_value=0, default_currency='CHF')


def refund_payment_form(request, payment_id):
    payment = get_object_or_404(Payment, pk=payment_id)
    if request.method == 'POST':
        form = RefundPaymentForm(request.POST)
        if form.is_valid():
            result = refund(
                amount=form.cleaned_data['amount'],
                payment_id=payment_id)
            # As confirmation we take the user to the edit page of the refund.
            return HttpResponseRedirect(reverse('admin:datatrans_refund_change', args=[result.id]))
    else:
        form = RefundPaymentForm(initial={'amount': payment.amount})

    return render(
        request,
        'admin/datatrans/form.html',
        {
            'title': 'Refund to {} {}, for original payment of {}'.format(payment.payment_method,
                                                                          payment.masked_card_number, payment.amount),
            'form': form,
            'opts': Payment._meta,  # Used to setup the navigation / breadcrumbs of the page
        }
    )


@admin.register(Payment)
class PaymentAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    list_display = [
        'transaction_id', 'created', 'success', 'client_ref', amount, 'payment_method', 'card_alias',
        'masked_card_number', expiry,
        'credit_card_country', 'error_code', 'refund_button']
    search_fields = [
        'transaction_id', 'created', 'expiry_date', 'payment_method', 'card_alias', 'masked_card_number', 'amount',
        'expiry_month', 'expiry_year', 'credit_card_country', 'client_ref', 'response_code',
        'authorization_code', 'acquirer_authorization_code', 'error_code', 'error_message', 'error_detail']
    list_filter = ['success', 'payment_method', 'credit_card_country',
                   ('amount_currency', admin.AllValuesFieldListFilter)]
    ordering = ['-created']

    readonly_fields = ['created', 'modified', 'refund_button']

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            url(
                r'^(?P<payment_id>[0-9a-f-]+)/refund/$',
                self.admin_site.admin_view(refund_payment_form),
                name='datatrans_payment_refund',
            ),
        ]
        return my_urls + urls

    def refund_button(self, obj):
        if obj.success:
            return format_html('<a class="button" href="{}">Refund</a>',
                               reverse('admin:datatrans_payment_refund', args=[obj.pk]))
        else:
            return '-'

    refund_button.short_description = 'Refund'  # type: ignore


@admin.register(Refund)
class RefundAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    readonly_fields = ['created', 'modified']
    list_display = [
        'transaction_id', 'payment_transaction_id', 'created', 'success', 'client_ref', amount, 'error_code',
        'error_message']
    list_display_links = ['transaction_id', 'payment_transaction_id']

    search_fields = [
        'transaction_id', 'payment_transaction_id', 'created', 'client_ref', 'amount', 'response_code',
        'authorization_code', 'acquirer_authorization_code', 'error_code', 'error_message', 'error_detail']
    list_filter = ['success', ('amount_currency', admin.AllValuesFieldListFilter)]
    ordering = ['-created']



from collections import namedtuple
from moneyed import Money
from structlog import get_logger

from .xml_helpers import money_to_amount_and_currency
from ..config import sign_web, web_merchant_id

logger = get_logger()

PaymentParameters = namedtuple('PaymentParameters', 'merchant_id amount currency refno sign use_alias')


def build_payment_parameters(amount: Money, client_ref: str) -> PaymentParameters:
    """
    Builds the parameters needed to present the user with a datatrans payment form.

    :param amount: The amount and currency we want the user to pay
    :param client_ref: A unique reference for this payment
    :return: The parameters needed to display the datatrans form
    """
    merchant_id = web_merchant_id
    amount, currency = money_to_amount_and_currency(amount)
    refno = client_ref
    sign = sign_web(merchant_id, amount, currency, refno)

    parameters = PaymentParameters(
        merchant_id=merchant_id,
        amount=amount,
        currency=currency,
        refno=refno,
        sign=sign,
        use_alias=False,
    )

    logger.info('build-payment-parameters', parameters=parameters)

    return parameters


def build_register_credit_card_parameters(client_ref: str) -> PaymentParameters:
    """
    Builds the parameters needed to present the user with a datatrans form to register a credit card.
    Contrary to a payment form, datatrans will not show an amount.

    :param client_ref: A unique reference for this alias capture.
    :return: The parameters needed to display the datatrans form
    """

    amount = 0
    currency = 'CHF'  # Datatrans requires this value to be filled, so we use this arbitrary currency.
    merchant_id = web_merchant_id
    refno = client_ref
    sign = sign_web(merchant_id, amount, currency, refno)

    parameters = PaymentParameters(
        merchant_id=merchant_id,
        amount=amount,
        currency=currency,
        refno=refno,
        sign=sign,
        use_alias=True,
    )

    logger.info('building-payment-parameters', parameters=parameters)

    return parameters



from django.dispatch import Signal

alias_registration_done = Signal(providing_args=['instance', 'success'])
payment_by_user_done = Signal(providing_args=['instance', 'success'])
payment_with_alias_done = Signal(providing_args=['instance', 'success'])
refund_done = Signal(providing_args=['instance', 'success'])



import calendar
import uuid

from datetime import date
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from djmoney.models.fields import MoneyField

from .signals import alias_registration_done, payment_by_user_done, payment_with_alias_done, refund_done

CLIENT_REF_FIELD_SIZE = 18

expiry_month_validators = [MinValueValidator(1), MaxValueValidator(12)]
expiry_year_validators = [MinValueValidator(0), MaxValueValidator(99)]


def compute_expiry_date(two_digit_year: int, month: int) -> date:
    year = 2000 + two_digit_year
    _, last_day_of_month = calendar.monthrange(year, month)
    return date(year=year, month=month, day=last_day_of_month)


class TransactionBase(models.Model):
    """"
    All the fields that are common to the different transaction types.
    Some of the fields declared here are redefined in subclasses:
    For instance expiry_month and expiry_year are optional in this base class,
    but in an alias registration those fields are mandatory.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    merchant_id = models.CharField(db_index=True, max_length=255)
    transaction_id = models.CharField(unique=True, max_length=18, blank=True, null=True)
    client_ref = models.CharField(db_index=True, max_length=18)
    amount = MoneyField(max_digits=12, decimal_places=2, default_currency='CHF')
    request_type = models.CharField(max_length=3, blank=True)

    expiry_month = models.IntegerField(null=True, blank=True, validators=expiry_month_validators)
    expiry_year = models.IntegerField(null=True, blank=True, validators=expiry_year_validators)

    expiry_date = models.DateField(null=True, blank=True)  # A field in the database so we can search for expired cards
    credit_card_country = models.CharField(db_index=True, max_length=3, blank=True)

    success = models.BooleanField()

    # If it's a success
    response_code = models.CharField(max_length=4, blank=True)
    response_message = models.CharField(max_length=255, blank=True)
    authorization_code = models.CharField(max_length=255, blank=True)
    acquirer_authorization_code = models.CharField(max_length=255, blank=True)

    # If it's an error
    error_code = models.CharField(max_length=7, blank=True)
    error_message = models.CharField(max_length=255, blank=True)
    error_detail = models.CharField(max_length=255, blank=True)
    acquirer_error_code = models.CharField(max_length=255, blank=True)

    def save(self, *args, **kwargs):
        if self.expiry_year is not None and self.expiry_month is not None:
            self.expiry_date = compute_expiry_date(two_digit_year=self.expiry_year, month=self.expiry_month)
        super().save(*args, **kwargs)

    def _send_signal(self, signal):
        signal.send(sender=None, instance=self, success=self.success)

    class Meta:
        abstract = True

    def __str__(self):
        return '{} {} ({})'.format(
            self.__class__.__name__,
            self.transaction_id,
            'successful' if self.success else 'failed'
        )

    def __repr__(self):
        c = self.__class__
        return '<{}.{} object at {}: {}>'.format(c.__module__, c.__name__, hex(id(self)), vars(self))


class AliasRegistration(TransactionBase):
    card_alias = models.CharField(db_index=True, max_length=20, blank=True)
    masked_card_number = models.CharField(max_length=255, blank=True)
    payment_method = models.CharField(db_index=True, max_length=3)
    expiry_month = models.IntegerField(validators=expiry_month_validators)
    expiry_year = models.IntegerField(validators=expiry_year_validators)

    def send_signal(self):
        self._send_signal(alias_registration_done)

    def __str__(self):
        return 'Alias: {} {}'.format(self.payment_method, self.masked_card_number)


class Payment(TransactionBase):
    transaction_id = models.CharField(unique=True, max_length=18)
    card_alias = models.CharField(db_index=True, max_length=20, blank=True)
    masked_card_number = models.CharField(max_length=255, blank=True)
    payment_method = models.CharField(db_index=True, max_length=3, blank=True)

    def send_signal(self):
        if self.card_alias:
            self._send_signal(payment_with_alias_done)
        else:
            self._send_signal(payment_by_user_done)


class Refund(TransactionBase):
    payment_transaction_id = models.CharField(db_index=True, max_length=18)

    def send_signal(self):
        self._send_signal(refund_done)



from xml.etree.ElementTree import Element, SubElement, tostring

from defusedxml.ElementTree import fromstring
from moneyed import Money
import requests
from structlog import get_logger

from .xml_helpers import money_to_amount_and_currency, parse_money
from ..config import datatrans_authorize_url, mpo_merchant_id, sign_mpo
from ..models import AliasRegistration, Payment

logger = get_logger()


def pay_with_alias(amount: Money, alias_registration_id: str, client_ref: str) -> Payment:
    """
    Charges money using datatrans, given a previously registered credit card alias.

    :param amount: The amount and currency we want to charge
    :param alias_registration_id: The alias registration to use
    :param client_ref: A unique reference for this charge
    :return: a Payment (either successful or not)
    """
    if amount.amount <= 0:
        raise ValueError('Pay with alias takes a strictly positive amount')

    alias_registration = AliasRegistration.objects.get(pk=alias_registration_id)

    logger.info('paying-with-alias', amount=amount, client_ref=client_ref,
                alias_registration=alias_registration)

    request_xml = build_pay_with_alias_request_xml(amount, client_ref, alias_registration)

    logger.info('sending-pay-with-alias-request', url=datatrans_authorize_url, data=request_xml)

    response = requests.post(
        url=datatrans_authorize_url,
        headers={'Content-Type': 'application/xml'},
        data=request_xml)

    logger.info('processing-pay-with-alias-response', response=response.content)

    charge_response = parse_pay_with_alias_response_xml(response.content)
    charge_response.save()
    charge_response.send_signal()

    return charge_response


def build_pay_with_alias_request_xml(amount: Money, client_ref: str, alias_registration: AliasRegistration) -> bytes:
    merchant_id = mpo_merchant_id
    client_ref = client_ref

    root = Element('authorizationService')
    root.set('version', '3')

    body = SubElement(root, 'body')
    body.set('merchantId', merchant_id)

    transaction = SubElement(body, 'transaction')
    transaction.set('refno', client_ref)

    request = SubElement(transaction, 'request')

    amount, currency = money_to_amount_and_currency(amount)
    SubElement(request, 'amount').text = str(amount)
    SubElement(request, 'currency').text = currency
    SubElement(request, 'aliasCC').text = alias_registration.card_alias
    SubElement(request, 'expm').text = str(alias_registration.expiry_month)
    SubElement(request, 'expy').text = str(alias_registration.expiry_year)
    SubElement(request, 'reqtype').text = 'CAA'
    SubElement(request, 'sign').text = sign_mpo(merchant_id, amount, currency, client_ref)

    return tostring(root, encoding='utf8')


def parse_pay_with_alias_response_xml(xml: bytes) -> Payment:
    body = fromstring(xml).find('body')
    status = body.get('status')
    transaction = body.find('transaction')
    trx_status = transaction.get('trxStatus')
    request = transaction.find('request')

    common_attributes = dict(
        merchant_id=body.get('merchantId'),
        card_alias=request.find('aliasCC').text,
        expiry_month=int(request.find('expm').text),
        expiry_year=int(request.find('expy').text),
        request_type=request.find('reqtype').text,
        client_ref=transaction.get('refno'),
        amount=parse_money(request)
    )

    if status == 'accepted' and trx_status == 'response':
        response = transaction.find('response')

        transaction_id = response.find('uppTransactionId').text
        masked_card_number = response.find('maskedCC').text
        return_customer_country = response.find('returnCustomerCountry').text

        response_code = response.find('responseCode').text
        response_message = response.find('responseMessage').text
        authorization_code = response.find('authorizationCode').text
        acquirer_authorization_code = response.find('acqAuthorizationCode').text

        d = dict(
            success=True,
            transaction_id=transaction_id,
            masked_card_number=masked_card_number,
            credit_card_country=return_customer_country,
            response_code=response_code,
            response_message=response_message,
            authorization_code=authorization_code,
            acquirer_authorization_code=acquirer_authorization_code,
        )
        d.update(common_attributes)
    else:
        error = transaction.find('error')

        d = dict(
            success=False,
            error_code=error.find('errorCode').text,
            error_message=error.find('errorMessage').text,
            error_detail=error.find('errorDetail').text,
            acquirer_error_code=text_or_else(error.find('acqErrorCode')),
            transaction_id=error.find('uppTransactionId').text,
            credit_card_country=text_or_else(error.find('returnCustomerCountry')),
        )
        d.update(**common_attributes)
    return Payment(**d)


def text_or_else(element, else_value='') -> str:
    if element is not None:
        return element.text
    else:
        return else_value
